/*
 Copyright 2015 Nicolas Melot

 This file is part of Drake-merge.

 Drake-merge is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Drake-merge is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Drake-merge. If not, see <http://www.gnu.org/licenses/>.

*/


#include <stdlib.h>
#include <math.h>

#include <drake.h>
#include <drake/link.h>
#include <drake/eval.h>
#include <string.h>

#include "sort.h"

#define INCLUDE_INITIAL_SORT 1

// Time calculation variables
static char *input_filename;

#if 10
#define debug(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%p\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

static int
greater(const void *a, const void *b)
{
	return *(int*)a - *(int*)b;
}

int
drake_init(task_t *task, void* aux)
{
	drake_platform_time_get(init[task->id -1]);
	return 1;
}

int
drake_start(task_t *task)
{
	// Record the starting time
	drake_platform_time_get(start[task->id - 1]);

	// Mark this task as executed
	execute[task->id - 1] = 1;
	core[task->id - 1] = drake_platform_core_id();

	// The task is always ready
	return 1;
}

int
drake_run(task_t *task)
{
	drake_platform_time_get(run[task->id - 1]);
	link_t *left_link, *right_link, *parent_link;

	if(pelib_map_next(string, link_tp)(pelib_map_next(string, link_tp)(pelib_map_begin(string, link_tp)(task->pred))) != pelib_map_end(string, link_tp)(task->pred) ||
		pelib_map_next(string, link_tp)(pelib_map_begin(string, link_tp)(task->succ)) != pelib_map_end(string, link_tp)(task->succ))
	{
		// Cannot merge, terminate immediately
		return 1;
	}

	// Fetch links
	left_link = pelib_map_read(string, link_tp)(pelib_map_find(string, link_tp)(task->pred, "left")).value;
	right_link = pelib_map_read(string, link_tp)(pelib_map_find(string, link_tp)(task->pred, "right")).value;
	parent_link = pelib_map_read(string, link_tp)(pelib_map_find(string, link_tp)(task->succ, "output")).value;

	// Fetch buffer addresses
	size_t left_size = 0, right_size = 0, parent_size = 0;
	int *left, *right, *parent;

	parent = pelib_cfifo_writeaddr(int)(parent_link->buffer, &parent_size, NULL);

	// Do stuff only if the parent can accept new input
	if(parent_size > 0)
	{
		left = pelib_cfifo_peekaddr(int)(left_link->buffer, 0, &left_size, NULL);
		right = pelib_cfifo_peekaddr(int)(right_link->buffer, 0, &right_size, NULL);

		// Start merge if both links have inputs
		if(left_size > 0 && right_size > 0)
		{
			int *left_ptr = left;
			int *right_ptr = right;
			int *parent_ptr = parent;

			// g++ generates faster code than gcc
			merge(&left_ptr, left + left_size, &right_ptr, right + right_size, &parent_ptr, parent + parent_size);
	
			// Commit data consumed and written
			pelib_cfifo_discard(int)(left_link->buffer, left_ptr - left);
			pelib_cfifo_discard(int)(right_link->buffer, right_ptr - right);
			pelib_cfifo_fill(int)(parent_link->buffer, parent_ptr - parent);
		}
		else
		{
			// If at least one input link has no input, then perhaps it's because the task terminated
			int left_killed = ((left_link->prod == NULL) ? 1 : left_link->prod->status >= TASK_KILLED);
			if(left_size == 0 && left_killed)
			{
				// Just copy from right input to parent
				size_t size = parent_size > right_size ? right_size : parent_size;
				memcpy(parent, right, sizeof(int) * size);
				pelib_cfifo_discard(int)(right_link->buffer, size);
				pelib_cfifo_fill(int)(parent_link->buffer, size);
			}

			int right_killed = ((right_link->prod == NULL) ? 1 : right_link->prod->status >= TASK_KILLED);
			if(right_size == 0 && right_killed)
			{
				// Just copy from right input to parent
				size_t size = parent_size > left_size ? left_size : parent_size;
				memcpy(parent, left, sizeof(int) * size);
				pelib_cfifo_discard(int)(left_link->buffer, size);
				pelib_cfifo_fill(int)(parent_link->buffer, size);
			}
		}
	}

	return drake_task_depleted(task);
}

int
drake_kill(task_t *task)
{
	drake_platform_time_get(killed[task->id - 1]);

	// Do nothing
	return 0;
}

int
drake_destroy(task_t *task)
{
	// Do nothing
	return 1;
}
