/*
 Copyright 2015 Nicolas Melot

 This file is part of Drake-merge.

 Drake-merge is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Drake-merge is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Drake-merge. If not, see <http://www.gnu.org/licenses/>.

*/


#include <stdlib.h>
#include <math.h>

#include <drake.h>
#include <drake/link.h>
#include <drake/eval.h>
#include <string.h>

#include "sort.h"

#define INCLUDE_INITIAL_SORT 1

#if 10
#define debug(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%p\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d:P%zu][%s] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

static array_t(int) *tmp;
static size_t total_size;
int check_output = 0;

#ifdef INPUT_READ
static char *filename;
#else
static unsigned int seed;
#endif

static int
greater(const void *a, const void *b)
{
	return *(int*)a - *(int*)b;
}

int
drake_init(task_t *task, void* aux)
{
	/* if task has no predecessor, then it creates two tasks
	for which it loads a portion of input and sort it, until it obtains
	two sorted subsequences that can be read from as if it was a fifo */

	// Record time of init
	drake_platform_time_get(init[task->id -1]);

	args_t *args = (args_t*)aux;

	if(args->argc > 0)
	{
#if INPUT_READ
		filename = ((args_t*)aux)->argv[0];
		tmp = pelib_array_preloadfilenamebinary(int)(filename);

		// Read input size
		if(tmp != NULL)
		{
			total_size = pelib_array_length(int)(tmp);
		}
		else
		{
			fprintf(stdout, "[%s:%s:%d:P%zu:%s] Cannot open input file \"%s\". Check application arguments.\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, filename);
		}

		// Enable result checking
		if(args->argc > 1)
		{
			check_output = 1;
		}
#else
		// Enable result checking
		if(args->argc > 2)
		{
			check_output = 1;
		}
		total_size = atoi(((args_t*)aux)->argv[0]);
		seed = atoi(((args_t*)aux)->argv[1]);
#endif
	}
	else
	{
#if INPUT_READ
		fprintf(stdout, "[%s:%s:%d:P%zu:%s] Missing file to read input from. Check application arguments.\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name);
#else
		fprintf(stdout, "[%s:%s:%d:P%zu:%s] Require size and random generator seed parameters from command line. Make sure aplication arguments are given in the form --application-args <size> <seed> --\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name);
#endif
	}

	// This will pass if the task is the root of the merging tree
	// but it is not also a leaf, that is, a single task. If it was
	// so, then the code above would have already added a successor link
	tmp = pelib_alloc_collection(array_t(int))(total_size);

	// Do nothing
	return 1;
}

int
drake_start(task_t *task)
{
	// Record teh starting time
	drake_platform_time_get(start[task->id - 1]);

	// Mark this task as executed
	execute[task->id - 1] = 1;
	core[task->id - 1] = drake_platform_core_id();

	// The task is always ready
	return 1;
}

int
drake_run(task_t *task)
{
	drake_platform_time_get(run[task->id - 1]);
	link_t *left_link, *right_link, *parent_link;

	if(pelib_map_next(string, link_tp)(pelib_map_next(string, link_tp)(pelib_map_begin(string, link_tp)(task->pred))) != pelib_map_end(string, link_tp)(task->pred))
	{
		// Cannot merge, terminate immediately
		return 1;
	}

	// Fetch links
	left_link = pelib_map_read(string, link_tp)(pelib_map_find(string, link_tp)(task->pred, "left")).value;
	right_link = pelib_map_read(string, link_tp)(pelib_map_find(string, link_tp)(task->pred, "right")).value;

	// Fetch buffer addresses
	size_t left_size = 0, right_size = 0, parent_size = 0;
	int *left, *right, *parent;

	parent_size = pelib_array_capacity(int)(tmp) - pelib_array_length(int)(tmp);
	parent = tmp->data + pelib_array_length(int)(tmp);

	left = pelib_cfifo_peekaddr(int)(left_link->buffer, 0, &left_size, NULL);
	right = pelib_cfifo_peekaddr(int)(right_link->buffer, 0, &right_size, NULL);

	// Start merge if both links have inputs
	if(left_size > 0 && right_size > 0)
	{
		int *left_ptr = left;
		int *right_ptr = right;
		int *parent_ptr = parent;

		// g++ generates faster code than gcc
		merge(&left_ptr, left + left_size, &right_ptr, right + right_size, &parent_ptr, parent + parent_size);

		// Commit data consumed and written
		pelib_cfifo_discard(int)(left_link->buffer, left_ptr - left);
		pelib_cfifo_discard(int)(right_link->buffer, right_ptr - right);
		tmp->length += (parent_ptr - parent);
	}
	else
	{
		// If at least one input link has no input, then perhaps it's because the task terminated
		int left_killed = ((left_link->prod == NULL) ? 1 : left_link->prod->status >= TASK_KILLED);
		if(left_size == 0 && left_killed)
		{
			// Just copy from right input to parent
			size_t size = parent_size > right_size ? right_size : parent_size;
			memcpy(parent, right, sizeof(int) * size);
			pelib_cfifo_discard(int)(right_link->buffer, size);
			tmp->length += size;
		}

		int right_killed = ((right_link->prod == NULL) ? 1 : right_link->prod->status >= TASK_KILLED);
		if(right_size == 0 && right_killed)
		{
			// Just copy from right input to parent
			size_t size = parent_size > left_size ? left_size : parent_size;
			memcpy(parent, left, sizeof(int) * size);
			pelib_cfifo_discard(int)(left_link->buffer, size);
			tmp->length += size;
		}
	}

	// Add the working time spent
	//drake_platform_time_get(stop_time);
	//drake_platform_time_substract(period, stop_time, start_time);
	//drake_platform_time_add(run[task->id - 1], run[task->id - 1], period);

	return drake_task_depleted(task);
}

int
drake_kill(task_t *task)
{
	drake_platform_time_get(killed[task->id - 1]);

	// Do nothing
	return 0;
}

int
drake_destroy(task_t *task)
{
	size_t i;
	//if(pelib_array_length(link_tp)(task->succ) == 0 && check_output != 0) // root node
	if(check_output != 0) // root node
	{
#if INPUT_READ
		array_t(int) *reference = pelib_array_loadfilenamebinary(int)(filename);
#else
		array_t(int) *reference = pelib_alloc_collection(array_t(int))(total_size);
		size_t num_leaf = (drake_task_number() + 1) / 2;
		size_t i;
		for(i = 0; i < total_size; i++)
		{
			int value = rand_r(&seed) % total_size;
			pelib_array_append(int)(reference, value);
		}
#endif
		
		if(pelib_array_length(int)(tmp) != pelib_array_length(int)(reference))
		{
			fprintf(stdout, "[%s:%s:%d:P%zu:%s] Different length of sorted (%zu) and reference arrays (%zu).\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, pelib_array_length(int)(tmp), pelib_array_length(int)(reference));
		}
		else
		{
			fprintf(stdout, "[%s:%s:%d:P%zu:%s] Input length OK.\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name);
			int sort_ok = 1;
			qsort((char*)reference->data, pelib_array_length(int)(reference), sizeof(int), greater);
			for(i = 0; i < pelib_array_length(int)(tmp); i++)
			{
				if(reference->data[i] != tmp->data[i])
				{
					sort_ok = 0;
					fprintf(stdout, "[%s:%s:%d:P%zu:%s] Different value at index %zu. Got %d, expected %d.\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name, i, tmp->data[i], reference->data[i]);
				}
			}
			if(sort_ok)
			{
				fprintf(stdout, "[%s:%s:%d:P%zu:%s] No difference between sorted and reference arrays.\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), task->name);
			}
		}
	}

	// Do nothing
	return 1;
}

